import json
import sys
from typing import ClassVar

from aibs_informatics_core.models.base.model import ModelBase
from aibs_informatics_core.utils.json import JSONObject

if sys.version_info < (3, 11):
    from typing_extensions import Self  # type: ignore[import-untyped]
else:
    from typing import Self

try:
    from pydantic import AliasGenerator, ConfigDict
    from pydantic import BaseModel as _PydanticBaseModel
    from pydantic.alias_generators import to_camel

except ModuleNotFoundError:  # pragma: no cover
    import types

    class _MissingPydantic(types.ModuleType):
        """Stub that raises a helpful error when any attribute is accessed."""

        __all__ = ()  # type: ignore

        def __getattr__(self, item):
            raise ImportError(
                "Optional dependency 'pydantic' is required for "
                "`aibs_informatics_core.models.base.PydanticBaseModel`. "
                "Install it with: pip install 'aibs-informatics-core[pydantic]'"
            )

    # Ensure subsequent `import pydantic` resolves to the stub so recursive
    # import attempts don’t continually re‑raise a generic ModuleNotFoundError.
    sys.modules.setdefault("pydantic", _MissingPydantic("pydantic"))

else:
    # --------------------------------------------------------------
    #                     PydanticModel
    # --------------------------------------------------------------

    class PydanticBaseModel(_PydanticBaseModel, ModelBase):
        """Base class for Pydantic models that can be serialized to/from JSON"""

        model_config: ClassVar[ConfigDict] = ConfigDict(
            populate_by_name=True,
            extra="ignore",
            alias_generator=AliasGenerator(
                # Use custom alias generators for validation and serialization
                # to ensure camelCase to snake_case conversion
                # and vice versa, depending on the context.
                validation_alias=to_camel,
                serialization_alias=to_camel,
            ),
        )

        @classmethod
        def from_dict(cls, data: JSONObject, **kwargs) -> Self:
            return cls.model_validate(
                data,
                **kwargs,
            )

        def to_dict(self, **kwargs) -> JSONObject:
            # Ensure None values are excluded by default to mirror DataClassJsonMixin settings
            exclude_none = kwargs.pop("exclude_none", True)
            mode = kwargs.pop("mode", "json")
            return self.model_dump(
                mode=mode,  # Use JSON serialization mode
                exclude_none=exclude_none,  # Exclude None values by default
                **kwargs,
            )

        @classmethod
        def from_json(cls, data: str, **kwargs) -> Self:
            return cls.from_dict(json.loads(data), **kwargs)



import logging
from typing import Optional

from aibs_informatics_core.executors.base import BaseExecutor
from aibs_informatics_core.utils.modules import load_type_from_qualified_name

logger = logging.getLogger(__name__)


def get_cli_parser():
    import argparse

    parser = argparse.ArgumentParser(description="Executor CLI")

    parser.add_argument(
        "--executor",
        dest="executor",
        required=True,
        help=(
            "executor class to run. Must be a fully qualified name of the executor class. "
            "e.g. aibs_informatics_aws_lambda.common.executor.BaseExecutor"
        ),
    )
    parser.add_argument(
        "--input",
        "--request",
        "-i",
        dest="request",
        required=True,
        help=(
            "input to executor. Can be a json string, json file, or S3 location. "
            "e.g. s3://bucket/key, /path/to/file.json, '{'foo': 'bar'}'"
        ),
    )
    parser.add_argument(
        "--output-location",
        "--response-location",
        "-o",
        dest="output_location",
        required=False,
        help=("optional response location to store response at. can be S3 or local file."),
    )
    return parser


def run_cli_executor(args: Optional[list[str]] = None):
    parsed_args = get_cli_parser().parse_args(args=args)

    executor_class = load_type_from_qualified_name(parsed_args.executor)

    if not issubclass(executor_class, BaseExecutor):
        raise ValueError(f"Executor class {executor_class} is not a subclass of BaseExecutor")

    executor_class.run_executor(parsed_args.request, parsed_args.output_location)


if __name__ == "__main__":  # pragma: no cover
    run_cli_executor()



import re
from dataclasses import dataclass
from typing import Callable, ClassVar, Dict, List, Match, Pattern, Type, TypeVar, Union

from aibs_informatics_core.collections import ValidatedStr
from aibs_informatics_core.models.demand_execution.resolvables import Resolvable


class JobParamEnvName(ValidatedStr):
    regex_pattern: ClassVar[Pattern] = re.compile(r"([_a-zA-Z][_a-zA-Z0-9]*)")

    def __new__(cls, value):
        normalized_value = value.replace("-", "_").replace(".", "_").upper()
        obj = super().__new__(cls, normalized_value)
        obj._raw = value
        return obj


class JobParamRef(ValidatedStr):
    regex_pattern: ClassVar[Pattern] = re.compile(r"\$\{([_a-zA-Z][_a-zA-Z0-9]*)\}")

    @property
    def raw_envname(self) -> str:
        return self.get_match_groups()[0]

    @property
    def envname(self) -> str:
        return JobParam.as_envname(self.raw_envname)

    @classmethod
    def from_name(cls, name) -> "JobParamRef":
        return JobParamRef(f"${{{JobParamEnvName(name)}}}")

    @classmethod
    def replace_references(
        cls, value: str, reference_replacement: Union[Dict[str, str], Callable[[Match], str]]
    ) -> str:
        if isinstance(reference_replacement, dict):
            reference_value_map = reference_replacement

            def replace_reference(match: Match) -> str:
                ref = JobParamRef(match.group(0))
                k = JobParam.as_envname(ref.envname)
                if k not in reference_value_map:
                    raise ValueError(f"Could not fill {ref}, no value in reference value map.")
                return reference_value_map[k]

            return JobParamRef.suball(string=value, repl=replace_reference)
        else:
            return JobParamRef.suball(string=value, repl=reference_replacement)


T = TypeVar("T")


@dataclass
class JobParam:
    name: str
    value: str

    def __hash__(self) -> int:
        return hash(self.name + str(self.value))

    @property
    def envname(self) -> JobParamEnvName:
        """Formats the name of the input to be an environment variable.
        For example: reference-path --> REFERENCE_PATH
        Replaces dashes with underscores and converts to upper case.
        """
        return self.as_envname(self.name)

    @property
    def envname_reference(self) -> JobParamRef:
        """Formats the name of the input to be used as an environment
        variable in a string. For example:
            reference-path --> ${REFERENCE_PATH}
        """
        return self.as_envname_reference(self.envname)

    def update_environment(self, environment: Dict[str, str], overwrite: bool = False):
        if (
            not overwrite
            and self.envname in environment
            and environment[self.envname] != self.value
        ):
            raise ValueError(
                f"Environment already contains {self.envname} with a different value."
            )
        environment[self.envname] = self.value

    @classmethod
    def as_envname(cls, name: str) -> JobParamEnvName:
        return JobParamEnvName(name)

    @classmethod
    def as_envname_reference(cls, name) -> JobParamRef:
        return JobParamRef.from_name(cls.as_envname(name))

    def find_references(self) -> List[JobParamRef]:
        if isinstance(self.value, str):
            return JobParamRef.findall(self.value)
        return []

    def replace_references(
        self, reference_replacement: Union[Dict[str, str], Callable[[Match], str]]
    ):
        self.value = JobParamRef.replace_references(
            value=self.value, reference_replacement=reference_replacement
        )


RP = TypeVar("RP", bound="ResolvableJobParam")


@dataclass
class ResolvableJobParam(JobParam):
    remote_value: str

    def __hash__(self) -> int:
        return hash(self.name + self.value + self.remote_value)

    def find_references(self) -> List[JobParamRef]:
        return super().find_references() + JobParamRef.findall(self.remote_value)

    def replace_references(
        self, reference_replacement: Union[Dict[str, str], Callable[[Match], str]]
    ):
        super().replace_references(reference_replacement=reference_replacement)
        self.remote_value = JobParamRef.replace_references(
            value=self.remote_value, reference_replacement=reference_replacement
        )

    @classmethod
    def from_resolvable(cls: Type[RP], name: str, resolvable: Resolvable) -> RP:
        return cls(name=name, value=resolvable.local, remote_value=resolvable.remote)


@dataclass
class DownloadableJobParam(ResolvableJobParam):
    def __hash__(self) -> int:
        return hash(self.name + self.value + self.remote_value)


@dataclass
class UploadableJobParam(ResolvableJobParam):
    def __hash__(self) -> int:
        return hash(self.name + self.value + self.remote_value)



from dataclasses import dataclass
from typing import Any, ClassVar, Dict, Optional, Set

import marshmallow as mm

from aibs_informatics_core.models.base.model import SchemaModel


@dataclass
class DemandResourceRequirements(SchemaModel):
    invalid_constraints: ClassVar[Set] = {0, None}  # remove None and 0 entries

    gpu: Optional[int] = None
    memory: Optional[int] = None
    vcpus: Optional[int] = None

    @classmethod
    @mm.post_dump
    def _filter_invalid_constraints(cls, data: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        return {k: v for k, v in data.items() if v not in cls.invalid_constraints}



__all__ = [
    "ArchiveType",
    "extract_archive",
    "make_archive",
    "move_path",
    "copy_path",
    "remove_path",
    "get_path_size_bytes",
    "get_path_hash",
    "find_paths",
    "get_path_with_root",
    "strip_path_root",
    "PathLock",
    "CannotAcquirePathLockError",
]

import errno
import fcntl
import hashlib
import logging
import os
import re
import shutil
import tarfile
import tempfile
import zipfile
from collections import deque
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Dict, List, Literal, Optional, Pattern, Sequence, Union, cast

from aibs_informatics_core.utils.decorators import deprecated
from aibs_informatics_core.utils.os_operations import find_all_paths

ArchiveFile = Union[tarfile.TarFile, zipfile.ZipFile]


logger = logging.getLogger(__name__)


ArchiveFormat = Literal[
    "tar",
    "bztar",
    "gztar",
    "xztar",
    "zip",
]


class ArchiveType(Enum):
    # https://docs.python.org/3/library/shutil.html#shutil.get_archive_formats
    TAR = "tar"
    TAR_BZ = "bztar"
    TAR_GZ = "gztar"
    TAR_XZ = "xztar"
    ZIP = "zip"

    @property
    def archive_format(self) -> ArchiveFormat:
        return cast(ArchiveFormat, self.value)

    def is_archive_type(self, path: Path) -> bool:
        try:
            return self == self.from_path(path)
        except Exception:
            return False

    @classmethod
    def is_archive(cls, path: Path) -> bool:
        try:
            cls.from_path(path)
        except Exception:
            return False
        else:
            return True

    @classmethod
    def from_path(cls, path: Path) -> "ArchiveType":
        if not path.exists() or path.is_dir():
            raise ValueError(f"Path {path} does not exist or is a directory")
        if tarfile.is_tarfile(path):
            compression_type = cls._get_compression_type(path)
            if compression_type == "gz":
                return ArchiveType.TAR_GZ
            elif compression_type == "bz":
                return ArchiveType.TAR_BZ
            elif compression_type == "xz":
                return ArchiveType.TAR_XZ
            else:
                return ArchiveType.TAR
        elif zipfile.is_zipfile(path):
            return ArchiveType.ZIP
        else:
            raise ValueError("Cannot infer type")

    @classmethod
    def _get_compression_type(cls, path: Path) -> Optional[Literal["gz", "bz", "xz", "zip"]]:
        file_sign_mapping: Dict[bytes, Literal["gz", "bz", "xz", "zip"]] = {
            b"\x1f\x8b\x08": "gz",
            b"\x42\x5a\x68": "bz",
            b"\xfd\x37\x7a\x58\x5a\x00": "xz",
            b"\x50\x4b\x03\x04": "zip",
        }

        max_len = max(len(x) for x in file_sign_mapping)
        with open(path, "rb") as f:
            file_start = f.read(max_len)
        for file_sign, filetype in file_sign_mapping.items():
            if file_start.startswith(file_sign):
                return filetype
        return None


def extract_archive(source_path: Path, destination_path: Optional[Path] = None) -> Path:
    """Untar/unzip data batch into a dedicate folder
    Example: batch_of_samples.tar.gz -> batch_of_samples

    Args:
        source_path (Path): archived batch data
        destination_path (Optional[Path]): Optional destination path for extracted data.
            If none, then source path name used with extension removed.

    Raises:
        RuntimeError: If extraction fails

    Returns:
        Path: path to the untarred data
    """
    archive_type = ArchiveType.from_path(source_path)

    extracted_path = destination_path or Path(tempfile.mkdtemp())

    shutil.unpack_archive(
        filename=str(source_path),
        extract_dir=str(extracted_path),
        format=archive_type.archive_format,
    )
    return extracted_path


def make_archive(
    source_path: Path,
    destination_path: Optional[Path] = None,
    archive_type: Union[ArchiveType, ArchiveFormat] = ArchiveType.TAR_GZ,
) -> Path:
    """tar/zip data batch from a folder
    Example: batch_of_samples -> batch_of_samples.tar.gz

    Args:
        source_path (Path): folder of data to archive
        destination_path (Optional[Path]): Optional destination path for archived file.
            If none, then tmp file is created and used

    Raises:
        RuntimeError: If archiving operation fails

    Returns:
        Path: path to the untarred data
    """

    archive_path = destination_path or Path(tempfile.mktemp())
    if not isinstance(archive_type, ArchiveType):
        archive_type = ArchiveType(archive_type)

    try:
        actual_archive_path = Path(
            shutil.make_archive(
                base_name=str(archive_path.with_suffix("")),
                root_dir=str(source_path),
                base_dir=None,
                format=archive_type.archive_format,
            )
        )

        # shutil.make_archive takes base name and appends the appropriate suffix.
        # This means we can't say "write at X". We must move results, but only if different.
        if actual_archive_path != archive_path:
            move_path(actual_archive_path, archive_path)
        return archive_path
    except Exception as e:
        raise ValueError(f"Error extracting file {source_path}. [{e}]") from e


def find_filesystem_boundary(starting_path: Path) -> Path:
    """Given some starting Path, determine the nearest filesystem boundary (mount point).
    If no mount is found, then this function will return the first parent directory PRIOR
    to the filesystem anchor.

    >>> find_filesystem_boundary(Path("/allen/scratch/aibstemp"))
    PosixPath('/allen/scratch')

    >>> find_filesystem_boundary(Path("/tmp/random_file.txt"))
    PosixPath('/tmp')

    Args:
        starting_path (Path): The starting Path

    Raises:
        RuntimeError: If the provided starting_path cannot resolve to a real existing path

    Returns:
        Path: The path of the nearest filesystem boundary OR the first parent directory prior
            to the filesystem anchor (example anchors: "/", "c:\\")
    """
    current_path = starting_path.resolve()
    while current_path.parent != Path(current_path.anchor):
        if current_path.is_mount():
            break
        current_path = current_path.parent

    if current_path.exists():
        return current_path
    else:
        raise OSError(f"Could not find a real filesystem boundary for: {str(starting_path)}")


def move_path(source_path: Path, destination_path: Path, exists_ok: bool = False):
    """Alias to simple mv command from one path to another

    Args:
        source_path (Path): source path
        destination_path (Path): destination path
        exists_ok (bool, optional): if true, overwrites destination. Defaults to False.

    """
    if destination_path.exists():
        if not exists_ok:
            raise ValueError(f"Cannot move path to {destination_path}. destination exists!")
        remove_path(destination_path)

    shutil.move(str(source_path), destination_path)


def copy_path(source_path: Path, destination_path: Path, exists_ok: bool = False):
    """Copies path from source to destination

    Args:
        source_path (Path): source path
        destination_path (Path): destination path
        exists_ok (bool, optional): if true, overwrites destination. Defaults to False.

    """
    if source_path.is_file():
        if destination_path.exists():
            if not exists_ok:
                raise ValueError(f"Cannot copy path to {destination_path}. destination exists!")
            remove_path(destination_path)
        shutil.copy(source_path, destination_path)
    else:
        shutil.copytree(source_path, destination_path, dirs_exist_ok=exists_ok)


def remove_path(path: Path, ignore_errors: bool = True):
    """Removes the contents at the path, if it exists"""
    try:
        if path.exists():
            if path.is_dir():
                shutil.rmtree(path, ignore_errors=ignore_errors)
            else:
                os.remove(path)
    except FileNotFoundError as e:
        # Ignore errors if requested
        if not ignore_errors:
            raise e
    except OSError as ose:
        # Ignore errors if requested
        if not ignore_errors:
            raise ose
        else:
            logger.warning(f"Failed to remove path {path}. Reason: {ose}")
            return


def get_path_size_bytes(path: Path) -> int:
    size_bytes = 0
    file_paths = deque(find_all_paths(path, include_dirs=False, include_files=True))
    while file_paths:
        file_path = file_paths.popleft()
        path = Path(file_path)
        try:
            size_bytes += path.stat().st_size
        except FileNotFoundError as e:
            logger.warning(f"File at {path} does not exist anymore. Reason: {e}")
            continue
        except OSError as ose:
            if ose.errno == errno.ESTALE:
                logger.warning(f"{ose} raised for {file_path}.")
                try:
                    if Path(str(path)).exists():
                        logger.warning(f"Adding {path} to end of list to check later.")
                        file_paths.append(file_path)
                except (FileNotFoundError, OSError) as fall_back_e:
                    logger.warning(f"Failed to check if path {path} exists: {fall_back_e}")
                    continue
            else:
                logger.error(f"Unexpected error raised for {path}. Reason: {ose}")
                raise ose
    return size_bytes


@deprecated("Please use `generate_path_hash` from `aibs_informatics_core.utils.hashing` instead")
def get_path_hash(
    path: Union[Path, str],
    includes: Optional[Sequence[Union[Pattern, str]]] = None,
    excludes: Optional[Sequence[Union[Pattern, str]]] = None,
) -> str:
    """Generate the hash based on files found under a given path.

    Args:
        path (str): path to compute a hash
        includes (Sequence[str], optional): list of regex patterns to include. Defaults to all.
        excludes (Sequence[str], optional): list of regex patterns to exclude. Defaults to None.

    Returns:
        str: hash value
    """
    from aibs_informatics_core.utils.hashing import generate_file_hash

    paths_to_hash = find_paths(root=path, include_dirs=False, includes=includes, excludes=excludes)

    path_hash = hashlib.sha256()
    for path in paths_to_hash:
        path_hash.update(generate_file_hash(path).encode("utf-8"))

    return path_hash.hexdigest()


def find_paths(
    root: Union[str, Path],
    include_dirs: bool = True,
    include_files: bool = True,
    includes: Optional[Sequence[Union[Pattern, str]]] = None,
    excludes: Optional[Sequence[Union[Pattern, str]]] = None,
) -> List[str]:
    """Find paths that match criteria

    Args:
        root (Union[str, Path]): root path
        include_dirs (bool, optional): whether to include directories. Defaults to True.
        include_files (bool, optional): whether to include files. Defaults to True.

        includes (Sequence[str], optional): list of regex patterns to include. Defaults to all.
        excludes (Sequence[str], optional): list of regex patterns to exclude. Defaults to None.

    Returns:
        List[str]: list of paths matching criteria
    """

    paths_to_return = []
    paths = find_all_paths(root, include_dirs=include_dirs, include_files=include_files)

    if not includes and not excludes:
        return paths

    include_patterns = [re.compile(include) for include in includes or [r".*"]]
    exclude_patterns = [re.compile(exclude) for exclude in excludes or []]

    for path in paths:
        target_path = path  # if check_absolute_path else strip_path_root(root=root, path=path)
        # First check exclude patterns
        for exclude_pattern in exclude_patterns:
            if exclude_pattern.fullmatch(target_path):
                break
        else:
            # Now check include patterns
            for include_pattern in include_patterns:
                if include_pattern.fullmatch(target_path):
                    paths_to_return.append(path)
                    break
    return paths_to_return


def get_path_with_root(path: Union[str, Path], root: Union[str, Path]) -> str:
    orig_path = path
    root = Path(root)
    path = Path(path)
    if path.is_relative_to(root):
        return str(orig_path)
    rel_path = path.relative_to("/") if path.is_absolute() else path
    path_with_root = os.path.normpath(root / rel_path)
    return path_with_root


def strip_path_root(path: Union[str, Path], root: Optional[Union[str, Path]] = None) -> str:
    """Strip the root from the path if path is absolute

    Args:
        path (Union[str, Path]): path to strip root from
        root (Optional[Union[str, Path]], optional): optionally specify root.
            If no root specified, uses "/".

    Returns:
        str: a relative path
    """
    root = Path(root) if root is not None else Path("/")
    path = Path(path)
    rel_path = path.relative_to(root) if path.is_absolute() else path
    return os.path.normpath(rel_path)


class CannotAcquirePathLockError(Exception):
    pass


@dataclass
class PathLock:
    """
    A context manager for acquiring and releasing locks on a file or directory path.

    If lock_root is provided, a lock file will be created in that directory with the name of the hash of the path.
    If lock_root is not provided, a lock file with the same name as the path and a .lock extension will be created.

    Providing an explicit lock root is useful if you dont want processes to read the lock file
    from the same directory as the file being locked.

    Attributes:
        path (Union[str, Path]): The path to the file.
        lock_root (Optional[Union[str, Path]]): The root directory for lock files. If provided, a
            lock file will be created in this directory with the name of the hash of the path.
            Otherwise, a lock file with the same name as the path and a .lock extension
            will be created. Defaults to None.
    """  # noqa: E501

    path: Union[str, Path]
    lock_root: Optional[Union[str, Path]] = None
    raise_if_locked: bool = False

    def __post_init__(self):
        # If lock root is provided, then create a lock file in that directory
        # with the name of the hash of the path. Otherwise, create a lock file
        # with the same name as the path with a .lock extension.
        if self.lock_root:
            lock_file_name = f"{hashlib.sha256(str(self.path).encode()).hexdigest()}.lock"
            self._lock_path = Path(self.lock_root) / lock_file_name
        else:
            self._lock_path = Path(f"{self.path}.lock")
        self._lock_file = None
        logger.info(f"Created {self} with {self._lock_path} lock file")

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, exec_type, exec_val, exec_tb):
        self.release()

    def acquire(self):
        logger.info("Acquiring lock...")
        try:
            self._lock_path.parent.mkdir(parents=True, exist_ok=True)
            self._lock_file = open(self._lock_path, "w")
            op = fcntl.LOCK_EX
            if self.raise_if_locked:
                op |= fcntl.LOCK_NB
            fcntl.flock(self._lock_file, op)
            self._lock_file.write(f"{datetime.now().timestamp()}")
            logger.info("Lock acquired!")
        except Exception as e:
            msg = f"Could not acquire lock! Reason: {e}"
            logger.error(msg)
            raise CannotAcquirePathLockError(msg) from e

    def release(self):
        logger.info("Releasing lock...")

        if self._lock_file and not self._lock_file.closed:
            try:
                fcntl.flock(self._lock_file, fcntl.LOCK_UN)
                self._lock_file.close()
            except Exception as e:
                logger.warning(f"Lock file doesn't exist. Skipping fcntl.flock and close: {e}")
        else:
            logger.warning("Strange! lock file already closed. not calling fcntl.flock")
        logger.info("Removing lock file")
        remove_path(self._lock_path)

        logger.info("Lock released!")


# ---------------
# Helpers


@deprecated("Please use `generate_file_hash` from `aibs_informatics_core.utils.hashing` instead")
def sha256sum(filename: str, bufsize: int = 128 * 1024) -> str:
    """

    https://stackoverflow.com/a/70215084/4544508

    Args:
        filename (str): file to hash
        bufsize (int, optional): buffer size. Defaults to 128*1024.

    Returns:
        str: hash value of file
    """
    h = hashlib.sha256()
    buffer = bytearray(bufsize)
    buffer_view = memoryview(buffer)
    with open(filename, "rb", buffering=0) as f:
        while True:
            n = f.readinto(buffer_view)  # type: ignore
            if not n:
                break
            h.update(buffer_view[:n])
    return h.hexdigest()



__all__ = [
    "S3PathStats",
    "S3Path",
    "S3PathPlaceholder",
    "S3URI",
    "S3BucketName",
    "S3BucketNamePlaceholder",
    "S3KeyPrefix",
    "S3Key",
    "S3KeyPlaceholder",
    "S3StorageClass",
    "S3StorageClassStr",
    "S3TransferRequest",
    "S3CopyRequest",
    "S3TransferResponse",
    "S3CopyResponse",
    "S3UploadRequest",
    "S3DownloadRequest",
    "S3UploadResponse",
    "S3RestoreStatus",
    "S3RestoreStatusEnum",
]

import re
import sys
from dataclasses import dataclass
from datetime import datetime
from functools import cached_property
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    Pattern,
    Protocol,
    TypedDict,
    TypeVar,
    Union,
)
from urllib.parse import quote

from aibs_informatics_core.exceptions import ValidationError

if sys.version_info >= (3, 11):
    from typing import NotRequired

import marshmallow as mm
from dateutil import parser as date_parser  # type: ignore[import-untyped]

from aibs_informatics_core.collections import OrderedStrEnum, ValidatedStr
from aibs_informatics_core.models.base import CustomStringField, EnumField

if TYPE_CHECKING:  # pragma: no cover
    # from mypy_boto3_s3.service_resource import Object as S3_Object
    class S3_Object(Protocol):
        storage_class: Optional[str]

else:
    S3_Object = object


if sys.version_info >= (3, 11):

    class BucketAndKey(TypedDict):
        Bucket: str
        Key: str
        VersionId: NotRequired[str]

else:  # pragma: no cover

    class _BucketAndKeyOpt(TypedDict, total=False):
        VersionId: str

    class _BucketAndKeyReq(TypedDict):
        Bucket: str
        Key: str

    class BucketAndKey(_BucketAndKeyReq, _BucketAndKeyOpt):
        pass


@dataclass
class S3PathStats:
    last_modified: datetime
    size_bytes: int
    object_count: Optional[int]

    def __getitem__(self, key):
        return super().__getattribute__(key)


# --- Placeholder (tight & brace-safe; adjust char class as you like) ---
PLACEHOLDER_PATTERN = r"(?:\$\{[A-Za-z0-9._\-\[\]]+\})"

NORMAL_CHARS = r"a-zA-Z0-9!_.*'()\-"
SPECIAL_CHARS = r"&$@=;:+,? "

# S3 Key Pattern Pattern: strictly no placeholders
# NOTE: S3 keys can be empty strings, so we allow that here
S3_KEY_PATTERN_STR_NO_VARS = rf"(?:[{NORMAL_CHARS}{SPECIAL_CHARS}]+/?)*"

KEY_RUN = rf"[{NORMAL_CHARS}{SPECIAL_CHARS}]+"  # non-slash run
KEY_TOKEN = rf"(?:{KEY_RUN}|{PLACEHOLDER_PATTERN})"  # run OR placeholder
KEY_SEG = rf"{KEY_TOKEN}+"  # one or more tokens per segment

# Mixed tokens where segments are separated by slashes, and can be empty (for trailing slash)
S3_KEY_PATTERN_STR_VARS = rf"(?:/(?={KEY_SEG}))?{KEY_SEG}(?:/{KEY_SEG})*/?|/?"

# S3 Bucket Name Pattern

BUCKET_RUN = r"[A-Za-z0-9.-]+"  # no slash
BUCKET_TOKEN = rf"(?:{BUCKET_RUN}|{PLACEHOLDER_PATTERN})"

# Strict bucket pattern
S3_BUCKET_NAME_PATTERN_STR_NO_VARS = r"(?:[A-Za-z0-9][A-Za-z0-9\-.]{1,61}[A-Za-z0-9])"

# Mixed tokens where the ends are alnum OR placeholder; handles 'bucket-${x}', '${x}-bucket', etc.
S3_BUCKET_NAME_PATTERN_MIXED = (
    rf"(?:"
    rf"(?:[A-Za-z0-9]|{PLACEHOLDER_PATTERN})"
    rf"(?:{BUCKET_TOKEN})*"
    rf"(?:[A-Za-z0-9]|{PLACEHOLDER_PATTERN})"
    rf")"
)

# Full bucket pattern: plain literal OR mixed OR single placeholder
S3_BUCKET_NAME_PATTERN_STR_VARS = (
    rf"(?:{S3_BUCKET_NAME_PATTERN_STR_NO_VARS}|"
    rf"{S3_BUCKET_NAME_PATTERN_MIXED}|"
    rf"{PLACEHOLDER_PATTERN})"
)


# https://stackoverflow.com/a/58248645/4544508
class S3BucketName(ValidatedStr):
    regex_pattern: ClassVar[Pattern] = re.compile(S3_BUCKET_NAME_PATTERN_STR_NO_VARS)

    def __truediv__(self, __other: str) -> "S3Path":
        """Creates a S3Path

        Examples:
            >>> s3_uri = S3BucketName("my-bucket") / "my-key"
            >>> assert s3_uri == "s3://my-bucket/my-key"

            >>> another_s3_uri = S3BucketName("bucket1") / S3Path("s3://bucket2/key2")
            >>> assert another_s3_uri == "s3://bucket1/key2"

        Args:
            __other (Union[str, S3Path]): The key or key of path to use for the S3Path

        Returns:
            S3Path: a new S3Path with the appended key using the `/` operator
        """

        if S3Path.is_valid(__other):
            __other = S3Path(__other).key
        return S3Path.build(bucket_name=self, key=__other)


class S3Key(ValidatedStr):
    regex_pattern: ClassVar[Pattern] = re.compile(S3_KEY_PATTERN_STR_VARS)

    @classmethod
    def _sanitize(cls, value: str, *args, **kwargs) -> str:
        value = str(value)
        value = _DOUBLE_SLASH_PATTERN.sub(r"\1", value)
        return value

    @property
    def components(self) -> List[str]:
        return self.split("/")

    def __rtruediv__(self, __other: str) -> "S3Key":
        """Creates a new S3 Key

        Args:
            __other (Union[str, S3Path]): The key to append to the end of this S3Path

        Returns:
            S3Path: a new S3Path with a new key
        """
        if isinstance(__other, str):
            prefix = __other.rstrip("/")
            return S3Key((prefix + "/" if prefix else "") + self)
        raise TypeError(f"{type(__other)} not supported for / operations with {type(self)}")


# https://stackoverflow.com/questions/58712045/regular-expression-for-amazon-s3-object-name
# NOTE: For now, this is is the same as S3Key. We made S3Key regex support an empty string
#       which was the only difference between the two.
class S3KeyPrefix(S3Key):
    pass


_DOUBLE_SLASH_PATTERN = re.compile(r"([^:]/)(/)+")


class S3Path(ValidatedStr):
    regex_pattern: ClassVar[Pattern] = re.compile(
        rf"^s3:\/\/({S3_BUCKET_NAME_PATTERN_STR_NO_VARS})(?:\/({S3_KEY_PATTERN_STR_NO_VARS}))?"
    )

    @classmethod
    def _sanitize(cls, value: str, *args, **kwargs) -> str:
        value = str(value)
        value = value[:3] + _DOUBLE_SLASH_PATTERN.sub(r"\1", value[3:])
        return value

    @cached_property
    def bucket(self) -> S3BucketName:
        return S3BucketName(self.get_match_groups()[0])

    @property
    def bucket_name(self) -> str:
        """Alias for bucket property"""
        return self.bucket

    @cached_property
    def key(self) -> S3Key:
        if key := self.get_match_groups()[1]:
            return S3Key(key)
        return S3Key("")

    @property
    def key_with_folder_suffix(self) -> str:
        return str(Path(self.key)) + "/"

    @property
    def name(self) -> str:
        return Path(self.key).name if self.key else ""

    @property
    def parent(self) -> "S3Path":
        # Handle root-level (just the bucket)
        if not self.key or self.key == "/":
            return S3Path.build(bucket_name=self.bucket_name, key="")

        # Strip trailing slash before splitting
        parent_key = "/".join(self.key.rstrip("/").split("/")[:-1]).rstrip("/") + "/"
        return S3Path.build(bucket_name=self.bucket_name, key=parent_key)

    @property
    def with_folder_suffix(self) -> "S3Path":
        return S3Path.build(bucket_name=self.bucket_name, key=self.key_with_folder_suffix)

    def has_folder_suffix(self) -> bool:
        return self.key.endswith("/")

    def as_dict(self) -> BucketAndKey:
        return BucketAndKey(Bucket=self.bucket, Key=self.key)

    def as_hosted_s3_url(self, aws_region: str) -> str:
        # TODO: need to encode special characters in key
        encoded_key = quote(self.key, safe="/")
        hosted_s3_url = f"https://{self.bucket}.s3.{aws_region}.amazonaws.com/{encoded_key}"
        return hosted_s3_url

    @classmethod
    def as_mm_field(cls) -> mm.fields.Field:
        return CustomStringField(S3Path)

    @classmethod
    def build(cls, bucket_name: str, key: str = "", **kwargs) -> "S3Path":
        """Build an `s3://` style URI given a bucket_name and key.

        There may be cases where the bucket_name or key is a placeholder
        (e.g. "${FILL_WITH_SOME_ENV_VAR}") in which case you must set allow_placeholders=True
        """
        bucket = S3BucketName(bucket_name, **kwargs)
        key = S3Key(key, **kwargs)
        return cls(f"s3://{bucket}/{key}", **kwargs)

    def __add__(self, __other: Union[str, "S3Path"]) -> "S3Path":
        """Appends a string or S3Path key to the end of this S3Path

        Examples:
            >>> s3_uri = S3Path("s3://my-bucket/my-key") + "-my-other-key"
            >>> assert s3_uri == "s3://my-bucket/my-key-my-other-key"

            >>> another_s3_uri = S3Path("s3://bucket1/key1") + S3Path("s3://bucket2/key2")
            >>> assert another_s3_uri == "s3://bucket1/key1key2"

        Args:
            __other (Union[str, S3Path]): The key to append to the end of this S3Path

        Returns:
            S3Path: a new S3Path with the appended key
        """
        if isinstance(__other, S3Path):
            __other = __other.key
        return S3Path(f"{self}{__other}")

    def __truediv__(self, __other: Union[str, "S3Path"]) -> "S3Path":
        """Appends a string or S3Path key to the end of this S3Path using the `/` operator

        Examples:
            >>> s3_uri = S3Path("s3://my-bucket/my-key") / "my-other-key"
            >>> assert s3_uri == "s3://my-bucket/my-key/my-other-key"

            >>> another_s3_uri = S3Path("s3://bucket1/key1") / S3Path("s3://bucket2/key2")
            >>> assert another_s3_uri == "s3://bucket1/key1/key2"

        Args:
            __other (Union[str, S3Path]): The key to append to the end of this S3Path

        Returns:
            S3Path: a new S3Path with the appended key using the `/` operator
        """
        if isinstance(__other, S3Path):
            __other = __other.key
        return S3Path(f"{self}/{__other}")

    def __floordiv__(self, __other: Union[str, "S3Path"]) -> "S3Path":
        """Creates a new S3Path by constructing a str or S3Path key with this bucket

        Examples:
            >>> s3_uri = S3Path("s3://my-bucket/my-key") // "my-other-key"
            >>> assert s3_uri == "s3://my-bucket/my-other-key"

            >>> another_s3_uri = S3Path("s3://bucket1/key1") // S3Path("s3://bucket2/key2")
            >>> assert another_s3_uri == "s3://bucket1/key2"

        Args:
            __other (Union[str, S3Path]): The key to append to the end of this S3Path

        Returns:
            S3Path: a new S3Path with a new key
        """
        if isinstance(__other, S3Path):
            __other = __other.key
        return S3Path.build(bucket_name=self.bucket, key=__other)


# Ensure backwards compatibility with old S3URI class name
S3URI = S3Path


class ConditionalPlaceholderStr(ValidatedStr):
    _placeholder_pattern: ClassVar[Pattern] = re.compile(PLACEHOLDER_PATTERN)

    def __new__(cls, value, *args, allow_placeholders: bool = False, **kwargs):
        return super().__new__(cls, value, *args, **kwargs)

    def __init__(self, *args, allow_placeholders: bool = False, **kwargs):
        self._allow_placeholders = allow_placeholders
        # TODO: This is to ensure backwards compatibility with the old `full_validate` kwarg
        #       Please remove this in the next major release
        if "full_validate" in kwargs:
            self._allow_placeholders = not kwargs.pop("full_validate")

    def _validate(self):
        super()._validate()
        if not self.allow_placeholders:
            if self.has_placeholder:
                raise ValidationError(
                    f"Placeholders are not allowed in {self} ({type(self)})"
                    f"with allow_placeholders={self.allow_placeholders}"
                )

    @property
    def allow_placeholders(self) -> bool:
        return self._allow_placeholders

    @cached_property
    def has_placeholder(self) -> bool:
        return bool(self._placeholder_pattern.search(self))


# https://stackoverflow.com/a/58248645/4544508
class S3BucketNamePlaceholder(ConditionalPlaceholderStr):
    regex_pattern: ClassVar[Pattern] = re.compile(S3_BUCKET_NAME_PATTERN_STR_VARS)

    def __truediv__(self, __other: str) -> "S3PathPlaceholder":
        """Creates a S3PathProxy

        Examples:
            >>> s3_uri = S3BucketName("my-bucket") / "my-key"
            >>> assert s3_uri == "s3://my-bucket/my-key"

            >>> another_s3_uri = S3BucketName("bucket1") / S3Path("s3://bucket2/key2")
            >>> assert another_s3_uri == "s3://bucket1/key2"

        Args:
            __other (Union[str, S3Path]): The key or key of path to use for the S3Path

        Returns:
            S3Path: a new S3Path with the appended key using the `/` operator
        """

        if S3PathPlaceholder.is_valid(__other):
            __other = S3PathPlaceholder(__other).key
        return S3PathPlaceholder.build(bucket_name=self, key=__other)


class S3KeyPlaceholder(ConditionalPlaceholderStr):
    regex_pattern: ClassVar[Pattern] = re.compile(S3_KEY_PATTERN_STR_VARS)

    @property
    def components(self) -> List[str]:
        return self.split("/")

    def __rtruediv__(self, __other: str) -> "S3KeyPlaceholder":
        """Creates a new S3 Key

        Args:
            __other (Union[str, S3Path]): The key to append to the end of this S3Path

        Returns:
            S3Path: a new S3Path with a new key
        """
        if isinstance(__other, str):
            prefix = __other.rstrip("/")
            return S3KeyPlaceholder((prefix + "/" if prefix else "") + self)
        raise TypeError(f"{type(__other)} not supported for / operations with {type(self)}")


class S3PathPlaceholder(ConditionalPlaceholderStr):
    regex_pattern: ClassVar[Pattern] = re.compile(
        rf"^s3:\/\/({S3_BUCKET_NAME_PATTERN_STR_VARS})(?:\/({S3_KEY_PATTERN_STR_VARS}))?"
    )

    @classmethod
    def _sanitize(cls, value: str, *args, **kwargs) -> str:
        value = value[:3] + _DOUBLE_SLASH_PATTERN.sub(r"\1", value[3:])
        return value

    @cached_property
    def bucket(self) -> S3BucketNamePlaceholder:
        return S3BucketNamePlaceholder(
            self.get_match_groups()[0], allow_placeholders=self.allow_placeholders
        )

    @property
    def bucket_name(self) -> str:
        """Alias for bucket property"""
        return self.bucket

    @cached_property
    def key(self) -> S3KeyPlaceholder:
        if key := self.get_match_groups()[1]:
            return S3KeyPlaceholder(key, allow_placeholders=self.allow_placeholders)
        return S3KeyPlaceholder("")

    @property
    def key_with_folder_suffix(self) -> str:
        return str(Path(self.key)) + "/"

    @property
    def name(self) -> str:
        return Path(self.key).name if self.key else ""

    @property
    def parent(self) -> "S3PathPlaceholder":
        parent_key = "/".join(self.key.rstrip("/").split("/")[:-1]).rstrip("/") + "/"
        return S3PathPlaceholder.build(
            bucket_name=self.bucket_name,
            key=parent_key,
            allow_placeholders=self.allow_placeholders,
        )

    @property
    def with_folder_suffix(self) -> "S3PathPlaceholder":
        return S3PathPlaceholder.build(
            bucket_name=self.bucket_name,
            key=self.key_with_folder_suffix,
            allow_placeholders=self.allow_placeholders,
        )

    def has_folder_suffix(self) -> bool:
        return self.key.endswith("/")

    def as_dict(self) -> BucketAndKey:
        return BucketAndKey(Bucket=self.bucket, Key=self.key)

    def as_hosted_s3_url(self, aws_region: str) -> str:
        # TODO: need to encode special characters in key
        encoded_key = quote(self.key, safe="/")
        hosted_s3_url = f"https://{self.bucket}.s3.{aws_region}.amazonaws.com/{encoded_key}"
        return hosted_s3_url

    @classmethod
    def as_mm_field(cls) -> mm.fields.Field:
        return CustomStringField(S3PathPlaceholder)

    @classmethod
    def build(
        cls, bucket_name: str, key: str = "", allow_placeholders: bool = False, **kwargs
    ) -> "S3PathPlaceholder":
        """Build an `s3://` style URI given a bucket_name and key.

        There may be cases where the bucket_name or key is a placeholder
        (e.g. "${FILL_WITH_SOME_ENV_VAR}") in which case you must set allow_placeholders=True
        """
        bucket = S3BucketNamePlaceholder(
            bucket_name, allow_placeholders=allow_placeholders, **kwargs
        )
        key = S3KeyPlaceholder(key, allow_placeholders=allow_placeholders, **kwargs)
        return cls(f"s3://{bucket}/{key}", allow_placeholders=allow_placeholders, **kwargs)

    def __add__(self, __other: Union[str, "S3PathPlaceholder", S3Path]) -> "S3PathPlaceholder":
        """Appends a string or S3Path key to the end of this S3Path

        Examples:
            >>> s3_uri = S3Path("s3://my-bucket/my-key") + "-my-other-key"
            >>> assert s3_uri == "s3://my-bucket/my-key-my-other-key"

            >>> another_s3_uri = S3Path("s3://bucket1/key1") + S3Path("s3://bucket2/key2")
            >>> assert another_s3_uri == "s3://bucket1/key1key2"

        Args:
            __other (Union[str, S3Path]): The key to append to the end of this S3Path

        Returns:
            S3Path: a new S3Path with the appended key
        """
        if isinstance(__other, (S3PathPlaceholder, S3Path)):
            __other = __other.key
        return S3PathPlaceholder(f"{self}{__other}", allow_placeholders=self.allow_placeholders)

    def __truediv__(self, __other: Union[str, S3Path, "S3PathPlaceholder"]) -> "S3PathPlaceholder":
        """Appends a string or S3Path key to the end of this S3Path using the `/` operator

        Examples:
            >>> s3_uri = S3Path("s3://my-bucket/my-key") / "my-other-key"
            >>> assert s3_uri == "s3://my-bucket/my-key/my-other-key"

            >>> another_s3_uri = S3Path("s3://bucket1/key1") / S3Path("s3://bucket2/key2")
            >>> assert another_s3_uri == "s3://bucket1/key1/key2"

        Args:
            __other (Union[str, S3Path]): The key to append to the end of this S3Path

        Returns:
            S3Path: a new S3Path with the appended key using the `/` operator
        """
        if isinstance(__other, (S3Path, S3PathPlaceholder)):
            __other = __other.key
        return S3PathPlaceholder(f"{self}/{__other}", allow_placeholders=self.allow_placeholders)

    def __floordiv__(
        self, __other: Union[str, S3Path, "S3PathPlaceholder"]
    ) -> "S3PathPlaceholder":
        """Creates a new S3Path by constructing a str or S3Path key with this bucket

        Examples:
            >>> s3_uri = S3Path("s3://my-bucket/my-key") // "my-other-key"
            >>> assert s3_uri == "s3://my-bucket/my-other-key"

            >>> another_s3_uri = S3Path("s3://bucket1/key1") // S3Path("s3://bucket2/key2")
            >>> assert another_s3_uri == "s3://bucket1/key2"

        Args:
            __other (Union[str, S3Path]): The key to append to the end of this S3Path

        Returns:
            S3Path: a new S3Path with a new key
        """
        if isinstance(__other, (S3Path, S3PathPlaceholder)):
            __other = __other.key
        return S3PathPlaceholder.build(
            bucket_name=self.bucket, key=__other, allow_placeholders=self.allow_placeholders
        )


T = TypeVar("T", S3Path, Path)
U = TypeVar("U", S3Path, Path)


@dataclass
class S3TransferRequest(Generic[T, U]):
    source_path: T
    destination_path: U


@dataclass
# class S3CopyRequest:
class S3CopyRequest(S3TransferRequest[S3Path, S3Path]):
    extra_args: Optional[Dict[str, Any]] = None


@dataclass
class S3TransferResponse:
    request: S3TransferRequest
    failed: bool = False
    reason: Optional[str] = None

    def __post_init__(self):
        if self.failed and not self.reason:
            raise ValueError(f"{self} must have a reason if failed.")


@dataclass
class S3CopyResponse:
    request: S3CopyRequest
    failed: bool = False
    reason: Optional[str] = None

    def __post_init__(self):
        if self.failed and not self.reason:
            raise ValueError(f"{self} must have a reason if failed.")


@dataclass
class S3UploadRequest(S3TransferRequest[Path, S3Path]):
    extra_args: Optional[Dict[str, Any]] = None


@dataclass
class S3DownloadRequest(S3TransferRequest[S3Path, Path]):
    pass


@dataclass
class S3UploadResponse:
    request: S3UploadRequest
    failed: bool = False
    reason: Optional[str] = None

    def __post_init__(self):
        if self.failed and not self.reason:
            raise ValueError(f"{self} must have a reason if failed.")


class S3RestoreStatusEnum(OrderedStrEnum):
    NOT_STARTED = "NOT_STARTED"
    IN_PROGRESS = "IN_PROGRESS"
    FINISHED = "FINISHED"


@dataclass
class S3RestoreStatus:
    restore_status: S3RestoreStatusEnum
    restore_expiration_time: Optional[datetime] = None

    @classmethod
    def from_raw_s3_restore_status(cls, raw_s3_restore_status: Optional[str]) -> "S3RestoreStatus":
        if raw_s3_restore_status is None:
            # Example of what boto3 s3.Object.restore property returns:
            # None
            return cls(restore_status=S3RestoreStatusEnum.NOT_STARTED)
        elif 'ongoing-request="true"' == raw_s3_restore_status:
            # Example of what boto3 s3.Object.restore property returns:
            # 'ongoing-request="true"'
            return cls(restore_status=S3RestoreStatusEnum.IN_PROGRESS)
        elif 'ongoing-request="false"' in raw_s3_restore_status:
            # Examples of what boto3 s3.Object.restore property returns:
            # 'ongoing-request="false", expiry-date="Fri, 21 Dec 2012 00:00:00 GMT"'
            # 'ongoing-request="false", expiry-date="Fri, 31 Mar 2023 00:00:00 GMT"'
            raw_time_str = raw_s3_restore_status.split("expiry-date=")[-1].strip('"')
            parsed_time = date_parser.parse(raw_time_str)
            return cls(
                restore_status=S3RestoreStatusEnum.FINISHED,
                restore_expiration_time=parsed_time,
            )
        else:
            raise RuntimeError(
                f"Could not parse the following raw_s3_restore_status: {raw_s3_restore_status}"
            )


class S3StorageClass(OrderedStrEnum):
    """OrderedStrEnum describing s3 storage classes from most to least accessible

    Convention: Deeper, less accessible storage classes are considered ">" than shallow, more
                accessible classes. (e.g. DEEP_ARCHIVE > STANDARD, GLACIER > GLACIER_IR, ...)
    """

    # Ordered from most accessible to least accessible
    # See: https://docs.aws.amazon.com/AmazonS3/latest/userguide/lifecycle-transition-general-considerations.html
    STANDARD = "STANDARD"
    STANDARD_IA = "STANDARD_IA"
    INTELLIGENT_TIERING = "INTELLIGENT_TIERING"
    ONEZONE_IA = "ONEZONE_IA"
    GLACIER_IR = "GLACIER_IR"
    GLACIER = "GLACIER"
    DEEP_ARCHIVE = "DEEP_ARCHIVE"
    # The following are special in that it's not easy to transition to or from them:
    OUTPOSTS = "OUTPOSTS"
    REDUCED_REDUNDANCY = "REDUCED_REDUNDANCY"

    @classmethod
    def from_boto_s3_obj(cls, s3_obj: S3_Object) -> "S3StorageClass":
        """Get S3StorageClass of an Boto3 S3_Object returned by s3.get_object"""
        if s3_obj.storage_class is None:
            return S3StorageClass.STANDARD
        else:
            return S3StorageClass(s3_obj.storage_class)

    @classmethod
    def list_archive_storage_classes(cls) -> List["S3StorageClass"]:
        """Storage classes that require a 'restore' operation to interact with the s3 object"""
        return [cls("GLACIER"), cls("DEEP_ARCHIVE")]

    @classmethod
    def list_transitionable_storage_classes(cls) -> List["S3StorageClass"]:
        return [
            cls("STANDARD"),
            cls("STANDARD_IA"),
            cls("INTELLIGENT_TIERING"),
            cls("ONEZONE_IA"),
            cls("GLACIER_IR"),
            cls("GLACIER"),
            cls("DEEP_ARCHIVE"),
        ]

    @classmethod
    def as_mm_field(cls) -> mm.fields.Field:
        return EnumField(S3StorageClass)


S3StorageClassStr = Literal[
    "STANDARD",
    "STANDARD_IA",
    "INTELLIGENT_TIERING",
    "ONEZONE_IA",
    "GLACIER_IR",
    "GLACIER",
    "DEEP_ARCHIVE",
    "OUTPOSTS",
    "REDUCED_REDUNDANCY",
]



__all__ = [
    "BooleanField",
    "CustomAwareDateTime",
    "CustomStringField",
    "DictField",
    "EnumField",
    "FloatField",
    "FrozenSetField",
    "IntegerField",
    "ListField",
    "MappingField",
    "NestedField",
    "PathField",
    "RawField",
    "StringField",
    "TupleField",
    "UnionField",
    "UUIDField",
]

import datetime as dt
import uuid
from collections import defaultdict
from enum import Enum
from inspect import isfunction
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import marshmallow as mm
from marshmallow import utils as mm_utils

from aibs_informatics_core.exceptions import ValidationError
from aibs_informatics_core.models.base.field_utils import FieldMetadataBuilder
from aibs_informatics_core.utils.json import JSON

T = TypeVar("T")
S = TypeVar("S", bound=str)
E = TypeVar("E", bound=Enum)

FieldMetadata = Dict[str, dict]

EncoderType = Callable[[T], JSON]
DecoderType = Callable[[JSON], T]


# --------------------------------------------------------------
#                  Custom Marshmallow Fields
# --------------------------------------------------------------


class EnumField(mm.fields.Field, Generic[E]):
    default_error_messages = {
        "invalid_value": "'{input}' is not a valid value of {obj_type}.",
        "not_enum": "'{input}' (type: {input_type}) is not an Enum type.",
    }

    def __init__(self, enum: Type[E], *args, **kwargs):
        self.enum_cls = enum
        super().__init__(*args, **kwargs)

    def _serialize(self, value, attr, obj, **kwargs) -> Optional[str]:
        if value is None:
            return None
        else:
            if isinstance(value, self.enum_cls):
                return value.value
            else:
                raise self.make_error(key="not_enum", input=value, input_type=type(value))

    def _deserialize(self, value, attr, data, **kwargs) -> Optional[E]:
        if value is None:
            return None
        else:
            try:
                return self.enum_cls(value)
            except ValueError as e:
                enum_name = self.enum_cls.__name__
                raise self.make_error(key="invalid_value", input=value, obj_type=enum_name) from e


class PathField(mm.fields.Field):
    def _serialize(self, value, attr, obj, **kwargs):
        return str(value) if isinstance(value, Path) else value

    def _deserialize(self, value, attr, data, **kwargs):
        return Path(value)


class FrozenSetField(mm.fields.List):
    def _serialize(self, value, *args, **kwargs):
        return super()._serialize(sorted(list(value)), *args, **kwargs)

    def _deserialize(self, *args, **kwargs):
        return frozenset(super()._deserialize(*args, **kwargs))


class CustomStringField(mm.fields.String, Generic[S]):
    default_error_messages = {
        "invalid_type": "'{input}' (type: {input_type}) is not a {expected_type} type!"
    }

    def __init__(self, str_cls: Type[S], *args, strict_mode: bool = False, **kwargs):
        """Field for subclassed string types.

        Deserialized values are constructed using subclass string type.

        By default, the serialization is relaxed on what inputs can be converted to strings
        (mirroring the default mm.fields.String behavior). i.e.
        If strict_mode is enabled, however, the

        Args:
            str_cls (Type[S]): _description_
            strict_mode (bool, optional): _description_. Defaults to False.
        """
        self.str_cls = str_cls
        self.strict_mode = strict_mode
        super().__init__(*args, **kwargs)

    def _serialize(self, value, attr, obj, **kwargs) -> Optional[str]:
        if value is None:
            return None
        elif not issubclass(type(value), self.str_cls):
            if not self._can_as(value):
                raise self._make_invalid_type_error(value=value)
            value = self.str_cls(value)
        return super()._serialize(value, attr, obj, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs) -> Optional[S]:
        deserialized_value = super()._deserialize(value=value, attr=attr, data=data, **kwargs)
        return self.str_cls(deserialized_value) if deserialized_value is not None else None

    def _make_invalid_type_error(self, value: Any) -> mm.ValidationError:
        return self.make_error(
            key="invalid_type",
            input=value,
            input_type=type(value),
            expected_type=self.str_cls,
        )

    def _can_as(self, value) -> bool:
        if self.strict_mode and not isinstance(value, self.str_cls):
            return False
        try:
            self.str_cls(value)
        except Exception:
            return False
        return True


class CustomAwareDateTime(mm.fields.AwareDateTime):
    """Version of marshmallow AwareDateTime field that allows an already deserialized
    (iso8601 str -> dt.datetime) AwareDateTime field to be passed through deserialization again
    (dt.datetime -> dt.datetime)

    This is necessary because the BaseModel may do a deserialization via the `load` method and
    then do another deserialization via the __setattr__ validation (which calls another
    deserialization to validate).

    Base implementation can be found here:
    https://marshmallow.readthedocs.io/en/stable/_modules/marshmallow/fields.html#AwareDateTime
    """

    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, dt.datetime):
            if not mm_utils.is_aware(value):
                if self.default_timezone is None:
                    raise self.make_error(
                        "invalid_awareness",
                        awareness=self.AWARENESS,
                        obj_type=self.OBJ_TYPE,
                    )
                return value.replace(tzinfo=self.default_timezone)
            return value
        return super()._deserialize(value, attr, data, **kwargs)


TypeOrTupleOfTypes = Union[Type[Any], Tuple[Type[Any], ...]]
UnionFieldsType = Tuple[
    Union[TypeOrTupleOfTypes, Callable[[], TypeOrTupleOfTypes]],
    Union[mm.fields.Field, Callable[[], mm.fields.Field]],
]


class UnionField(mm.fields.Field):
    default_error_messages = {
        "invalid_type": "'{input}' (type: {input_type}) is not one of {expected_types} type!",
        "unexpected_error": "Unexpected error occurred: {error}",
        "unsupported_serialize": (
            "'{input}' (type: {input_type}) was not serializable "
            "by any of the {expected_types} fields!: {error_messages}"
        ),
        "unsupported_deserialize": (
            "'{input}' (type: {input_type}) was not handled "
            "by any of the {expected_types} fields!: {error_messages}"
        ),
    }

    def __init__(self, union_fields: Sequence[UnionFieldsType], *args, **kwargs):
        self._raw_union_fields: Sequence[UnionFieldsType] = union_fields
        self._union_fields: Dict[Type[Any], List[mm.fields.Field]] = None  # type: ignore
        super().__init__(*args, **kwargs)

    @property
    def union_fields(self) -> Dict[Type[Any], List[mm.fields.Field]]:
        if self._union_fields is None:
            self._union_fields = defaultdict(list)
            for union_type, union_field in self._raw_union_fields:
                if isfunction(union_type):
                    union_type = cast(Callable[[], TypeOrTupleOfTypes], union_type)()
                union_types = union_type if isinstance(union_type, tuple) else (union_type,)
                if isfunction(union_field):
                    union_field = union_field()

                for union_type in union_types:
                    self._union_fields[cast(Type[Any], union_type)].append(
                        cast(mm.fields.Field, union_field)
                    )
        return self._union_fields

    def _serialize(self, value, attr, obj, **kwargs) -> Optional[str]:
        if value is None:
            return None
        input_type = type(value)
        errors: Dict[Type[Any], List[mm.ValidationError]] = defaultdict(list)
        for class_type, class_fields in self.union_fields.items():
            if issubclass(input_type, class_type):
                for class_field in class_fields:
                    try:
                        return class_field._serialize(value, attr=attr, obj=obj, **kwargs)
                    except mm.ValidationError as e:
                        errors[class_type].append(e)
                    except Exception as e:
                        print(e)
                        errors[class_type].append(self.make_error(key="unexpected_error", error=e))
        else:
            if len(errors):
                raise self.make_error(
                    key="unsupported_serialize",
                    input=value,
                    input_type=input_type,
                    expected_types=list(self.union_fields.keys()),
                    error_messages={
                        _.__name__: [e.normalized_messages() for e in e_list]
                        for _, e_list in errors.items()
                    },
                )

            raise self.make_error(
                key="invalid_type",
                input=value,
                input_type=input_type,
                expected_types=list(self.union_fields.keys()),
            )

    def _deserialize(self, value, attr, data, **kwargs) -> Optional[E]:
        if value is None:
            return None
        else:
            errors: Dict[Type[Any], List[mm.ValidationError]] = defaultdict(list)
            for class_type, class_fields in self.union_fields.items():
                for class_field in class_fields:
                    try:
                        return class_field._deserialize(
                            value=value, attr=attr, data=data, **kwargs
                        )
                    except (mm.ValidationError, ValidationError) as e:
                        if not isinstance(e, mm.ValidationError):
                            e = self.make_error(key="unexpected_error", error=e)
                        errors[class_type].append(e)
            else:
                raise self.make_error(
                    key="unsupported_deserialize",
                    input=value,
                    input_type=type(value),
                    expected_types=list(self.union_fields.values()),
                    error_messages={
                        _.__name__: [e.normalized_messages() for e in e_list]
                        for _, e_list in errors.items()
                    },
                )


BooleanField = mm.fields.Boolean
DictField = mm.fields.Dict
FloatField = mm.fields.Float
IntegerField = mm.fields.Integer
MappingField = mm.fields.Mapping
NestedField = mm.fields.Nested
ListField = mm.fields.List
RawField = mm.fields.Raw
StringField = mm.fields.String
TupleField = mm.fields.Tuple
UUIDField = mm.fields.UUID


# --------------------------------------------------------------
#                   Global Configurations
# --------------------------------------------------------------


FieldMetadataBuilder(mm_field=CustomAwareDateTime("iso8601")).add_to_global_config(
    dt.datetime, required=False, skip_mm_field=True
)

FieldMetadataBuilder(mm_field=UUIDField()).add_to_global_config(
    uuid.UUID, required=False, skip_mm_field=True
)



from __future__ import annotations

import datetime
import sys
from typing import Annotated

from aibs_informatics_core.utils.time import from_isoformat_8601

try:
    from pydantic import BeforeValidator, PlainSerializer

except ModuleNotFoundError:  # pragma: no cover
    import types

    class _MissingPydantic(types.ModuleType):
        """Stub that raises a helpful error when any attribute is accessed."""

        __all__ = ()  # type: ignore

        def __getattr__(self, item):
            raise ImportError(
                "Optional dependency 'pydantic' is required for "
                "`aibs_informatics_core.models.base.PydanticBaseModel`. "
                "Install it with: pip install 'aibs-informatics-core[pydantic]'"
            )

    # Ensure subsequent `import pydantic` resolves to the stub so recursive
    # import attempts don’t continually re‑raise a generic ModuleNotFoundError.
    sys.modules.setdefault("pydantic", _MissingPydantic("pydantic"))

else:

    def _parse_isoish_dt(v: str | int | float | datetime.datetime) -> datetime.datetime:
        """Handle `2025-04-30T07:00:00.0` and plain `datetime` values."""
        # Convert epoch‑milliseconds first
        if isinstance(v, (int, float)):
            return datetime.datetime.fromtimestamp(v / 1_000, datetime.timezone.utc)
        # Handle ISO 8601 date‑time strings

        # Clean up iso‑format strings that may include a fractional component like ".0"
        if isinstance(v, str):
            return from_isoformat_8601(v)  # This will raise ValueError if the format is incorrect

        # Already a datetime instance → return unchanged
        return v

    def _parse_date(v: str | datetime.date) -> datetime.date:
        """Handle date strings and plain `datetime.date` values."""
        if isinstance(v, str):
            return datetime.date.fromisoformat(v)
        return v

    # --------------------------------------------------------------
    #                     Pydantic Fields
    # --------------------------------------------------------------

    IsoDateTime = Annotated[
        datetime.datetime,
        BeforeValidator(_parse_isoish_dt),
        # ↓ this runs when .model_dump_json() or .model_dump(mode="json") is called
        PlainSerializer(
            lambda v: v.isoformat(),  # or v.isoformat(timespec="seconds") for no µs
            return_type=str,
            when_used="json",  # only affects JSON/dict output, not Python copy
        ),
    ]
    IsoDate = Annotated[
        datetime.date,
        BeforeValidator(_parse_date),
        PlainSerializer(
            lambda v: v.isoformat(),  # or v.isoformat(timespec="seconds") for no µs
            return_type=str,
            when_used="json",  # only affects JSON/dict output, not Python copy
        ),
    ]



__all__ = [
    "b64_decoded_str",
    "b64_encoded_str",
    "generate_file_hash",
    "generate_path_hash",
    "sha256_hexdigest",
    "urlsafe_b64_decoded_str",
    "urlsafe_b64_encoded_str",
    "uuid_str",
]

import hashlib
import json
import logging
import re
import uuid
from base64 import standard_b64decode, standard_b64encode, urlsafe_b64decode, urlsafe_b64encode
from pathlib import Path
from typing import List, Literal, Optional, Union

from aibs_informatics_core.utils.json import JSON
from aibs_informatics_core.utils.os_operations import find_all_paths

logger = logging.getLogger(__name__)


HashTypeStr = Literal["md5", "sha256", "sha1"]


def uuid_str(content: Optional[str] = None) -> str:
    """Get a UUID String, with option for using a seed to ensure determinism.

    Args:
        content (str, optional): A seed to use for determining UUID. Defaults to None.

    Returns:
        str: UUID appropriate string
    """
    if content is None:
        return str(uuid.uuid4())
    return str(uuid.uuid3(namespace=uuid.NAMESPACE_DNS, name=content))


def sha256_hexdigest(content: Optional[JSON] = None) -> str:
    """Create a SHA 256 Hex Digest string from optional content.

    If content is not provided, a unique Hex Digest is generated from UUID


    Args:
        content (JSON, optional): Input to base hexdigest off of. Defaults to None.

    Returns:
        str: a SHA 256 hex digest string.
    """
    if content is None:
        content = uuid_str()
    elif not isinstance(content, str):
        content = json.dumps(content, sort_keys=True)
    return hashlib.sha256(content.encode()).hexdigest()


def b64_decoded_str(encoded_str: str) -> str:
    """Decodes an encoded base64 string.

    Args:
        encoded_str (str): A string that has been previously encoded with base64

    Returns:
        str: a decoded base 64 string
    """
    try:
        return standard_b64decode(encoded_str.encode()).decode()
    except Exception as e:
        logger.error(e)
        logger.exception(e)
        raise e


def b64_encoded_str(decoded_str: str) -> str:
    """Encodes a string with base 64.

    Args:
        encoded_str (str): Any string

    Returns:
        str: an encoded base 64 string
    """
    return standard_b64encode(decoded_str.encode()).decode()


def urlsafe_b64_decoded_str(encoded_str: str) -> str:
    """Decodes an encoded base64 string.

    Args:
        encoded_str (str): A string that has been previously encoded with base64

    Returns:
        str: a decoded base 64 string
    """
    return urlsafe_b64decode(encoded_str.encode()).decode()


def urlsafe_b64_encoded_str(decoded_str: str) -> str:
    """Encodes a string with a URL SAFE version of base 64.

    Args:
        encoded_str (str): Any string

    Returns:
        str: an encoded base 64 string
    """
    return urlsafe_b64encode(decoded_str.encode()).decode()


def generate_path_hash(
    path: Union[str, Path],
    includes: Optional[List[str]] = None,
    excludes: Optional[List[str]] = None,
    hash_type: HashTypeStr = "sha256",
) -> str:
    """Generate a hash based on files found under a given path.

    Args:
        path (str): path to compute a hash
        includes (List[str], optional): list of regex patterns to include. Defaults to None.
        excludes (List[str], optional): list of regex patterns to exclude. Defaults to None.
        hash_type (Literal["md5", "sha256"], optional): type of hash to generate.
            Defaults to "sha256".

    Returns:
        str: hash value
    """
    paths = find_all_paths(path, include_dirs=False)
    include_patterns = [re.compile(include) for include in includes or [r".*"]]
    exclude_patterns = [re.compile(exclude) for exclude in excludes or []]

    paths_to_hash = []
    for path in paths:
        # First check exclude patterns
        for exclude_pattern in exclude_patterns:
            if exclude_pattern.fullmatch(path):
                break
        else:
            # Now check include patterns
            for include_pattern in include_patterns:
                if include_pattern.fullmatch(path):
                    paths_to_hash.append(path)
                    break
    path_hash = hashlib.new(hash_type)
    for path in paths_to_hash:
        path_hash.update(generate_file_hash(path, hash_type=hash_type).encode("utf-8"))

    return path_hash.hexdigest()


def generate_file_hash(
    filename: Union[str, Path], bufsize: int = 128 * 1024, hash_type: HashTypeStr = "sha256"
) -> str:
    """Generate a hash for a file

    https://stackoverflow.com/a/70215084/4544508

    Args:
        filename (str|Path): filepath to hash
        bufsize (int, optional): buffer size. Defaults to 128*1024.
        hash_type (Literal["md5", "sha256"], optional): type of hash to generate.
            Defaults to "sha256".

    Returns:
        str: hash value of file
    """
    filename = str(filename)
    h = hashlib.new(hash_type)

    buffer = bytearray(bufsize)
    buffer_view = memoryview(buffer)
    with open(filename, "rb", buffering=0) as f:
        while True:
            n = f.readinto(buffer_view)  # type: ignore
            if not n:
                break
            h.update(buffer_view[:n])
    return h.hexdigest()



import logging
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, cast

from aibs_informatics_core.models.aws.sfn import ExecutionArn
from aibs_informatics_core.models.base import SchemaModel, custom_field, pre_load
from aibs_informatics_core.models.base.custom_fields import (
    BooleanField,
    CustomStringField,
    DictField,
    EnumField,
    ListField,
    StringField,
)
from aibs_informatics_core.models.status import Status


@dataclass
class DemandExecutionMetadata(SchemaModel):
    user: Optional[str] = custom_field(default=None)
    arn: Optional[ExecutionArn] = custom_field(
        mm_field=CustomStringField(ExecutionArn), default=None
    )
    tags: Optional[Dict[str, str]] = custom_field(
        mm_field=DictField(StringField(), StringField(), allow_none=True),
        default=None,
    )
    notify_on: Optional[Dict[Status, bool]] = custom_field(
        mm_field=DictField(keys=EnumField(Status), values=BooleanField(), allow_none=True),
        default=None,
    )
    notify_list: Optional[List[str]] = custom_field(
        mm_field=ListField(StringField(), allow_none=True),
        default=None,
    )

    @property
    def tag(self) -> Optional[str]:
        """Return all tags as a comma separated string if available.

        Example:
            >>> metadata = DemandExecutionMetadata(tags={"key": "value"})
            >>> metadata.tag
            'key=value'
            >>> metadata = DemandExecutionMetadata(tags={"key": "key"})
            >>> metadata.tag
            'key'
            >>> metadata = DemandExecutionMetadata(tags={"flag": "flag", "key": "value"})
            >>> metadata.tag
            'flag,key=value'
            >>> metadata = DemandExecutionMetadata(tags={})
            >>> metadata.tag
            None

        Returns:
            Optional[str]: Comma separated string of tags or None if no tags are available.

        """
        if self.tags:
            return ",".join([f"{k}={v}" if k != v else k for k, v in self.tags.items()])
        return None

    @classmethod
    @pre_load
    def sanitize_tags(cls, data: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        """Sanitize tags and tag fields in the input data.

        If the `tag` field is present, it will be converted to a dictionary format. It supports
        both string and list formats. `tag` can include multiple tags separated by commas.

        By contrast, the `tags` field is expected to be a dictionary. If both `tag` and `tags` are
        present, they will be merged, with `tags` taking precedence.

        The `tag` field is deprecated and will be removed in a future version. It is recommended to
        use the `tags` field exclusively for tagging.

        TODO: Once we have datafixes for the places where demand execution metadata is used, we
        can remove this method and the `tag` field from the model. The `tags` field should be used
        exclusively.

        """
        norm_tags: dict[str, str] = {}
        if "tag" in data:
            tag_value = data.pop("tag")
            if isinstance(tag_value, (str, int)):
                tag_value = [tag_value]

            if isinstance(tag_value, list) and all([isinstance(_, (str, int)) for _ in tag_value]):
                norm_tags = dict(
                    [
                        (tag_part, tag_part)
                        if "=" not in tag_part
                        else cast(tuple[str, str], tuple(tag_part.split("=", 1)))
                        for _ in tag_value
                        for tag_part in str(_).split(",")
                    ]
                )
            else:
                logging.warning(
                    f"Invalid tag format ({type(tag_value)}): {tag_value}. "
                    "Expected str, int, or list of str/int. Ignoring."
                )

        if "tags" in data:
            tags = data["tags"]
            if isinstance(tags, dict):
                data["tags"] = norm_tags | tags
            else:
                raise ValueError(f"Invalid tags format ({type(tags)}): {tags}")
        elif norm_tags:
            data["tags"] = norm_tags
        return data



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, FrozenSet, Iterable, List, Optional, Union

from aibs_informatics_core.models.aws.s3 import S3URI
from aibs_informatics_core.models.base import (
    FrozenSetField,
    SchemaModel,
    StringField,
    custom_field,
)
from aibs_informatics_core.models.base.custom_fields import UnionField
from aibs_informatics_core.models.demand_execution.job_param import ResolvableJobParam


@dataclass
class ParamPair(SchemaModel):
    """SchemaModel for an input and output parameter pairs as described in demand execution

    The values for input and output should correspond to the parameter key in the demand execution

    Scenarios:
        1. input is not None and output is not None (input and output)
            This is used to represent a single job that has both inputs and outputs (Most common)
        2. input and output are both None (no inputs or outputs)
            This is used to represent a single job that has no inputs or outputs
        3. input is None and output is not None (output only)
            This is used to represent a single job that has only outputs
        4. input is not None and output is None (input only)
            This is used to represent a single job that has only inputs

    """

    input: Optional[str] = custom_field(mm_field=StringField(), default=None)
    output: Optional[str] = custom_field(mm_field=StringField(), default=None)

    @classmethod
    def from_set_pairs(cls, *values: ParamSetPair) -> List[ParamPair]:
        return [pair for value in values for pair in value.to_pairs()]

    @classmethod
    def from_sets(cls, inputs: Iterable[str], outputs: Iterable[str]) -> List[ParamPair]:
        if not inputs and not outputs:
            return []
        elif not inputs:
            return [ParamPair(output=output) for output in outputs]
        elif not outputs:
            return [ParamPair(input=input) for input in inputs]
        return [ParamPair(input=input, output=output) for input in inputs for output in outputs]


@dataclass
class ParamSetPair(SchemaModel):
    """SchemaModel for a set of input and output parameter pairs as described in demand execution

    The values for inputs and outputs should correspond to the parameter keys in the
    demand execution.

    Scenarios:
        1. inputs and outputs are not empty
            This accounts for set of related inputs and outputs
        2. inputs and outputs are both empty
            This is used to represent a single job that has no inputs or outputs
        3. inputs is empty and outputs is not empty
            This is used to represent a single job that has only outputs
        4. inputs is not empty and outputs is empty
            This is used to represent a single job that has only inputs
    """

    inputs: FrozenSet[str] = custom_field(
        mm_field=FrozenSetField(StringField), default_factory=frozenset
    )
    outputs: FrozenSet[str] = custom_field(
        mm_field=FrozenSetField(StringField), default_factory=frozenset
    )

    def __post_init__(self):
        if not isinstance(self.inputs, frozenset):
            self.inputs = frozenset(self.inputs)
        if not isinstance(self.outputs, frozenset):
            self.outputs = frozenset(self.outputs)

    def add_inputs(self, *inputs: str):
        self.inputs = self.inputs.union(inputs)

    def add_outputs(self, *outputs: str):
        self.outputs = self.outputs.union(outputs)

    def remove_inputs(self, *inputs: str):
        self.inputs = self.inputs.difference(inputs)

    def remove_outputs(self, *outputs: str):
        self.outputs = self.outputs.difference(outputs)

    def to_pairs(self) -> List[ParamPair]:
        return ParamPair.from_sets(inputs=self.inputs, outputs=self.outputs)

    @classmethod
    def from_pairs(cls, *pairs: ParamPair) -> List[ParamSetPair]:
        """Converts a list of ParamPairs to a list of ParamSetPairs

        This is useful for grouping ParamPairs by output

        Returns:
            List[ParamSetPair]: A list of ParamSetPairs
        """
        # group pairs only by outputs
        output_set_pairs: Dict[Union[str, None], ParamSetPair] = {}
        for pair in pairs:
            if pair.output not in output_set_pairs:
                output_set_pairs[pair.output] = ParamSetPair(
                    outputs=frozenset({pair.output}) if pair.output else frozenset()
                )
            if pair.input is not None:
                output_set_pairs[pair.output].add_inputs(pair.input)
        return list(output_set_pairs.values())


@dataclass(frozen=True)
class JobParamPair:
    """models an input and output resolved parameter pair as described in demand execution

    The values for input and outputs should account for both the parameter name and the
    remote location of the parameter value. The remote location can be a S3URI or other
    remote location.

    This captures all information about a parameter

    Scenarios:
        1. input is not None and output is not None (input and output)
            This is used to represent a single job that has both inputs and outputs (Most common)
        2. input and output are both None (no inputs or outputs)
            This is used to represent a single job that has no inputs or outputs
        3. input is None and output is not None (output only)
            This is used to represent a single job that has only outputs
        4. input is not None and output is None (input only)
            This is used to represent a single job that has only inputs
    """  # noqa: E501

    input: Optional[ResolvableJobParam] = None
    output: Optional[ResolvableJobParam] = None

    @classmethod
    def from_sets(
        cls, inputs: Iterable[ResolvableJobParam], outputs: Iterable[ResolvableJobParam]
    ) -> List[JobParamPair]:
        if not inputs and not outputs:
            return []
        elif not inputs:
            return [JobParamPair(output=output) for output in outputs]
        elif not outputs:
            return [JobParamPair(input=input) for input in inputs]
        return [JobParamPair(input=input, output=output) for input in inputs for output in outputs]

    @classmethod
    def from_set_pairs(cls, *values: JobParamSetPair) -> List[JobParamPair]:
        return [pair for value in values for pair in value.to_pairs()]


@dataclass
class JobParamSetPair:
    """models a set of input and output resolved parameter pairs as described in demand execution

    The values for inputs and outputs should account for both the parameter name and the remote location.
    The remote location can be a S3URI or other remote location.

    This captures all information about a parameter

    Scenarios:
        1. inputs and outputs are not empty
            This accounts for set of related inputs and outputs
        2. inputs and outputs are both empty
            This is used to represent a single job that has no inputs or outputs
        3. inputs is empty and outputs is not empty
            This is used to represent a single job that has only outputs
        4. inputs is not empty and outputs is empty
            This is used to represent a single job that has only inputs
    """  # noqa: E501

    inputs: FrozenSet[ResolvableJobParam] = field(default_factory=frozenset)
    outputs: FrozenSet[ResolvableJobParam] = field(default_factory=frozenset)

    def __post_init__(self):
        if not isinstance(self.inputs, frozenset):
            self.inputs = frozenset(self.inputs)
        if not isinstance(self.outputs, frozenset):
            self.outputs = frozenset(self.outputs)

    def add_inputs(self, *inputs: ResolvableJobParam):
        self.inputs = self.inputs.union(inputs)

    def add_outputs(self, *outputs: ResolvableJobParam):
        self.outputs = self.outputs.union(outputs)

    def remove_inputs(self, *inputs: ResolvableJobParam):
        self.inputs = self.inputs.difference(inputs)

    def remove_outputs(self, *outputs: ResolvableJobParam):
        self.outputs = self.outputs.difference(outputs)

    def to_pairs(self) -> List[JobParamPair]:
        return JobParamPair.from_sets(inputs=self.inputs, outputs=self.outputs)

    @classmethod
    def from_pairs(cls, *pairs: JobParamPair) -> List[JobParamSetPair]:
        """Converts a list of JobParamPairs to a list of JobParamSetPairs

        This is useful for grouping JobParamPairs by output

        Returns:
            List[JobParamSetPair]: A list of JobParamSetPairs
        """
        # group pairs only by outputs
        output_set_pairs: Dict[Union[ResolvableJobParam, None], JobParamSetPair] = {}
        for pair in pairs:
            if pair.output not in output_set_pairs:
                output_set_pairs[pair.output] = JobParamSetPair(
                    outputs=frozenset({pair.output}) if pair.output else frozenset()
                )
            if pair.input is not None:
                output_set_pairs[pair.output].add_inputs(pair.input)
        return list(output_set_pairs.values())


# ResolvableID = Union[S3URI, ...]
# TODO: need to add additional types.
ResolvableID = S3URI


@dataclass
class ResolvedParamSetPair(SchemaModel):
    """SchemaModel for a set of input and output resolved parameter pairs as described in demand execution

    This is the other side of the ParamSetPair coin. This is used to represent a set of
    inputs and outputs' remote locations. The remote location can be a S3URI or other
    remote location.

    This only captures the remote location of a parameter. It does not capture the parameter name.

    """  # noqa: E501

    inputs: FrozenSet[ResolvableID] = custom_field(
        mm_field=FrozenSetField(
            UnionField(
                [
                    (S3URI, S3URI.as_mm_field()),
                ]
            )
        ),
        default_factory=frozenset,
    )
    outputs: FrozenSet[ResolvableID] = custom_field(
        mm_field=FrozenSetField(
            UnionField(
                [
                    (S3URI, S3URI.as_mm_field()),
                ]
            )
        ),
        default_factory=frozenset,
    )

    def __post_init__(self):
        if not isinstance(self.inputs, frozenset):
            self.inputs = frozenset(self.inputs)
        if not isinstance(self.outputs, frozenset):
            self.outputs = frozenset(self.outputs)

    def add_inputs(self, *inputs: ResolvableID):
        self.inputs = self.inputs.union(inputs)

    def add_outputs(self, *outputs: ResolvableID):
        self.outputs = self.outputs.union(outputs)

    def remove_inputs(self, *inputs: ResolvableID):
        self.inputs = self.inputs.difference(inputs)

    def remove_outputs(self, *outputs: ResolvableID):
        self.outputs = self.outputs.difference(outputs)



__all__ = [
    "FieldMetadataBuilder",
    "FieldProps",
    "custom_field",
    "field_metadata",
]


from dataclasses import MISSING, Field, dataclass
from dataclasses import field as dataclasses_field
from enum import Enum
from typing import Any, Callable, Dict, Mapping, Optional, Type, TypeVar, cast, overload

import marshmallow as mm
from dataclasses_json import config, global_config
from typing_inspect import is_optional_type as _is_optional_type

from aibs_informatics_core.utils.json import JSON

T = TypeVar("T")
S = TypeVar("S", bound=str)
E = TypeVar("E", bound=Enum)

FieldMetadata = Dict[str, dict]

EncoderType = Callable[[T], JSON]
DecoderType = Callable[[JSON], T]


# --------------------------------------------------------------
#                     Field Metadata Builder
# --------------------------------------------------------------


@dataclass
class FieldMetadataBuilder:
    mm_field: Optional[mm.fields.Field] = None
    encoder: Optional[EncoderType] = None
    decoder: Optional[DecoderType] = None

    def build(self, required: Optional[bool] = None, **kwargs) -> FieldMetadata:
        mm_field = self.mm_field
        encoder = self.encoder
        decoder = self.decoder

        # set allow_none to !required IF required is explicitly specified (True/False).
        allow_none = not required if required is not None else required

        if mm_field is None:
            mm_field_cls = self.create_mm_field_class(encoder=encoder, decoder=decoder)
            mm_field_kwargs: dict[str, Any] = {}
            if required is not None:
                mm_field_kwargs["required"] = required
            mm_field = mm_field_cls(**mm_field_kwargs)
            if allow_none is not None:
                mm_field.allow_none = allow_none

        else:
            if required is not None:
                mm_field.required = required
                if allow_none is not None:
                    mm_field.allow_none = allow_none
            if encoder is None:
                encoder = self.create_encoder_from_mm_field(mm_field)
            if decoder is None:
                decoder = self.create_decoder_from_mm_field(mm_field)

        return config(encoder=encoder, decoder=decoder, mm_field=mm_field, **kwargs)

    def add_to_global_config(
        self,
        clazz: Type[Any],
        required: Optional[bool] = None,
        skip_mm_field: bool = False,
        skip_encoder: bool = False,
        skip_decoder: bool = False,
        **kwargs,
    ):
        dataclass_json_config = self.build(required=required, **kwargs)["dataclasses_json"]

        if "mm_field" in dataclass_json_config and not skip_mm_field:
            global_config.mm_fields[clazz] = dataclass_json_config["mm_field"]
        if "encoder" in dataclass_json_config and not skip_encoder:
            global_config.encoders[clazz] = dataclass_json_config["encoder"]
        if "decoder" in dataclass_json_config and not skip_decoder:
            global_config.decoders[clazz] = dataclass_json_config["decoder"]

    @classmethod
    def create_mm_field_class(
        cls, encoder: Optional[EncoderType], decoder: Optional[DecoderType]
    ) -> Type[mm.fields.Field]:
        """Creates a Marshmallow Field with optional encoder/decoder overrides

        Args:
            encoder (Optional[EncoderType]): Optional encoder function
            decoder (Optional[DecoderType]): Optional decoder function

        Returns:
            Type[mm.fields.Field]: _description_
        """

        class CustomField(mm.fields.Field):
            def _serialize(self, value, attr, obj, **kwargs):
                if encoder:
                    return encoder(value)
                return super()._serialize(value, attr, obj, **kwargs)

            def _deserialize(self, value, attr, data, **kwargs):
                if decoder:
                    return decoder(value)
                return super()._deserialize(value, attr, data, **kwargs)

        return CustomField

    @classmethod
    def create_encoder_from_mm_field(cls, mm_field: mm.fields.Field) -> EncoderType:
        def mm_field_encoder(value):
            return mm_field.serialize("", value, lambda _1, _2, _3: value)

        return cast(EncoderType, mm_field_encoder)

    @classmethod
    def create_decoder_from_mm_field(cls, mm_field: mm.fields.Field) -> DecoderType:
        def mm_field_decoder(value):
            return mm_field.deserialize(value)

        return cast(DecoderType, mm_field_decoder)


def field_metadata(
    mm_field: Optional[mm.fields.Field] = None,
    encoder: Optional[EncoderType] = None,
    decoder: Optional[DecoderType] = None,
    required: Optional[bool] = None,
    **kwargs,
) -> FieldMetadata:
    return FieldMetadataBuilder(mm_field=mm_field, encoder=encoder, decoder=decoder).build(
        required=required, **kwargs
    )


# --------------------------------------------------------------
#               Dataclass Utility Functions
# --------------------------------------------------------------


@dataclass
class FieldProps:
    field: Field

    def requires_init(self) -> bool:
        return not self.has_default()

    def is_optional_type(self) -> bool:
        return _is_optional_type(self.field.type)

    def has_default(self) -> bool:
        return not (self.field.default_factory is MISSING and self.field.default is MISSING)  # type: ignore[misc]


@overload
def custom_field(
    *,
    default: Optional[T],
    init: bool = True,
    repr: bool = True,
    hash: Optional[bool] = None,
    compare: bool = True,
    metadata: Optional[Mapping[Any, Any]] = None,
    mm_field: Optional[mm.fields.Field] = None,
    encoder: Optional[EncoderType] = None,
    decoder: Optional[DecoderType] = None,
) -> T: ...  # pragma: no cover


@overload
def custom_field(
    *,
    default_factory: Callable[[], T],
    init: bool = True,
    repr: bool = True,
    hash: Optional[bool] = None,
    compare: bool = True,
    metadata: Optional[Mapping[Any, Any]] = None,
    mm_field: Optional[mm.fields.Field] = None,
    encoder: Optional[EncoderType] = None,
    decoder: Optional[DecoderType] = None,
) -> T: ...  # pragma: no cover


@overload
def custom_field(
    *,
    init: bool = True,
    repr: bool = True,
    hash: Optional[bool] = None,
    compare: bool = True,
    metadata: Optional[Mapping[Any, Any]] = None,
    mm_field: Optional[mm.fields.Field] = None,
    encoder: Optional[EncoderType] = None,
    decoder: Optional[DecoderType] = None,
) -> Any: ...  # pragma: no cover


def custom_field(
    *,
    # Below values mirror those found in `dataclasses.field`
    default: Any = MISSING,
    default_factory: Callable[[], Any] = MISSING,  # type: ignore
    init: bool = True,
    repr: bool = True,
    hash: Optional[bool] = None,
    compare: bool = True,
    metadata: Optional[Mapping[Any, Any]] = None,
    # Following are custom
    mm_field: Optional[mm.fields.Field] = None,
    encoder: Optional[EncoderType] = None,
    decoder: Optional[DecoderType] = None,
) -> Any:
    """Convenience function for generating a Dataclass Field WITH Encoders/Decoders/ MM Fields

    Args:
        mm_field (mm.fields.Field, optional): _description_. Defaults to None.
        encoder (EncoderType, optional): _description_. Defaults to None.
        decoder (DecoderType, optional): . Defaults to None.
        default (Any, optional): Default value if not provided. Defaults to MISSING.
        default_factory (Callable[[], Any], optional): _description_. Defaults to MISSING.
        repr (bool, optional): _description_. Defaults to True.
        hash (Optional[bool], optional): _description_. Defaults to None.
        compare (bool, optional): _description_. Defaults to True.
        metadata (Optional[Mapping[Any, Any]], optional): _description_. Defaults to None.

    Returns:
        Field: _description_
    """
    required = init and not (default != MISSING or default_factory != MISSING)
    metadata = field_metadata(
        mm_field=mm_field,
        encoder=encoder,
        decoder=decoder,
        required=required,
        metadata=metadata,
    )
    if default != MISSING:
        return dataclasses_field(
            default=default,
            init=init,
            compare=compare,
            hash=hash,
            repr=repr,
            metadata=metadata,
        )
    elif default_factory != MISSING:
        return dataclasses_field(
            default_factory=default_factory,
            init=init,
            compare=compare,
            hash=hash,
            repr=repr,
            metadata=metadata,
        )
    else:
        return dataclasses_field(
            init=init,
            compare=compare,
            hash=hash,
            repr=repr,
            metadata=metadata,
        )


# to make it synonymous with dataclasses.field
field = custom_field



import re
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

from aibs_informatics_core.collections import ValidatedStr
from aibs_informatics_core.models.base import IntegerField, ListField, SchemaModel, custom_field


class JobName(ValidatedStr):
    regex_pattern = re.compile(r"([a-zA-Z0-9][\w_-]{0,127})")


@dataclass
class ResourceRequirements(SchemaModel):
    GPU: Optional[int] = custom_field(mm_field=IntegerField(strict=False), default=None)
    MEMORY: Optional[int] = custom_field(mm_field=IntegerField(strict=False), default=None)
    VCPU: Optional[int] = custom_field(mm_field=IntegerField(strict=False), default=None)


@dataclass
class KeyValuePairType(SchemaModel):
    Name: str
    Value: str


@dataclass
class ContainerDetail(SchemaModel):
    Image: str = custom_field()
    Environment: List[KeyValuePairType] = custom_field(
        mm_field=ListField(KeyValuePairType.as_mm_field())
    )
    ContainerInstanceArn: str = custom_field()
    TaskArn: str = custom_field()


@dataclass
class AttemptContainerDetail(SchemaModel):
    ContainerInstanceArn: Optional[str] = custom_field(default=None)
    TaskArn: Optional[str] = custom_field(default=None)
    ExitCode: Optional[int] = custom_field(default=None)
    Reason: Optional[str] = custom_field(default=None)
    LogStreamName: Optional[str] = custom_field(default=None)


@dataclass
class AttemptDetail(SchemaModel):
    Container: Optional[AttemptContainerDetail] = custom_field(
        mm_field=AttemptContainerDetail.as_mm_field(), default=None
    )
    StartedAt: Optional[int] = custom_field(default=None)
    StoppedAt: Optional[int] = custom_field(default=None)
    StatusReason: Optional[str] = custom_field(default=None)

    @property
    def duration(self) -> Optional[int]:
        if self.StoppedAt and self.StartedAt:
            return self.StoppedAt - self.StartedAt
        return None

    @property
    def container_instance_arn(self) -> Optional[str]:
        if self.Container:
            return self.Container.ContainerInstanceArn
        return None

    @property
    def container_task_arn(self) -> Optional[str]:
        if self.Container:
            return self.Container.TaskArn
        return None


@dataclass
class BatchJobDetail(SchemaModel):
    JobName: str = custom_field()
    JobId: str = custom_field()
    JobQueue: str = custom_field()
    Status: str = custom_field()
    StartedAt: int = custom_field()
    JobDefinition: str = custom_field()

    # Optional
    JobArn: Optional[str] = custom_field(default=None, repr=False)
    StatusReason: Optional[str] = custom_field(default=None, repr=False)
    Attempts: List[AttemptDetail] = custom_field(
        mm_field=ListField(AttemptDetail.as_mm_field()), default_factory=list, repr=False
    )
    Container: Optional[ContainerDetail] = custom_field(
        mm_field=ContainerDetail.as_mm_field(), default=None, repr=False
    )
    Parameters: Dict[str, str] = custom_field(default_factory=dict, repr=False)
    CreatedAt: Optional[int] = custom_field(default=None, repr=False)
    StoppedAt: Optional[int] = custom_field(default=None, repr=False)
    IsCancelled: Optional[bool] = custom_field(default=None, repr=False)
    IsTerminated: Optional[bool] = custom_field(default=None, repr=False)

    @property
    def duration(self) -> Optional[int]:
        if self.StoppedAt and self.StartedAt:
            return self.StoppedAt - self.StartedAt
        return None

    @property
    def container_name_and_tag(self) -> Tuple[str, Optional[str]]:
        if self.Container:
            (container, container_tag) = self.Container.Image.split(":", maxsplit=1)
            return (container, container_tag)
        return ("NotAvailable", None)

    @property
    def container_tag(self) -> Optional[str]:
        return self.container_name_and_tag[1]

    @property
    def container_environment(self) -> Dict[str, str]:
        return {
            env_var.Name: env_var.Value
            for env_var in (self.Container.Environment if self.Container else {})
        }

    @property
    def container_instance_arn(self) -> Optional[str]:
        if self.Container:
            return self.Container.ContainerInstanceArn
        return None

    @property
    def container_instance_arns(self) -> List[str]:
        return [
            _
            for _ in set(
                [
                    self.container_instance_arn,
                    *[
                        a.Container.ContainerInstanceArn
                        for a in self.Attempts or []
                        if a.Container
                    ],
                ]
            )
            if _ is not None
        ]

    @property
    def container_task_arn(self) -> Optional[str]:
        if self.Container:
            return self.Container.TaskArn
        return None



from dataclasses import dataclass
from typing import Optional, Union

from aibs_informatics_core.models.aws.iam import IAMRoleArn
from aibs_informatics_core.models.base import SchemaModel, custom_field
from aibs_informatics_core.models.base.custom_fields import (
    CustomStringField,
    StringField,
    UnionField,
)


@dataclass
class AWSBatchExecutionPlatform(SchemaModel):
    job_queue_name: str
    job_role: Optional[Union[str, IAMRoleArn]] = custom_field(
        mm_field=UnionField([(IAMRoleArn, CustomStringField(IAMRoleArn)), (str, StringField())]),
        default=None,
    )


# TODO: I would prefer to make ExecutionPlatform polymorphic, but datacalasses does not support it
#       For now, I will just make a
@dataclass
class ExecutionPlatform(SchemaModel):
    aws_batch: Optional[AWSBatchExecutionPlatform] = custom_field(
        mm_field=AWSBatchExecutionPlatform.as_mm_field(), default=None
    )
